 list
 ;---------------------------------------------------------------------
 ; CONTROL:
 ;	file		async.inc
 ;	author		wouter van ooijen <wf@xs4all.nl>
 ; 	version		1.0
 ; 	date		05-09-1997
 ; 	language	MPASM assembler for Microchip PIC processors
 ;      summary         various macros for asynchronous serial communication.
 ;      info            http://www.xs4all.nl/~wf/wouter/pic/wloader
 ;---------------------------------------------------------------------
 nolist
 ; Copyright (C) 2000 Wouter van Ooijen
 ;
 ; This application is free software; you can redistribute it and/or
 ; modify it under the terms of the GNU Library General Public
 ; License as published by the Free Software Foundation; either
 ; version 2 of the License, or (at your option) any later version.
 ;
 ; This library is distributed in the hope that it will be useful,
 ; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ; Library General Public License for more details.
 ; 
 ; You should have received a copy of the GNU Library General Public
 ; License along with this library; if not, write to the
 ; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 ; Boston, MA  02111-1307, USA.
 ;---------------------------------------------------------------------
 listing -1

 with delay.inc
 with convert.inc

;
; Macro for 8/N/1 asychronous serial sending with busy delays.
; Must be instantiated with the choosen 
; port, buffer, bit, baudrate and polarity.
; Requires:
; - xtal must be defined
; - macro delay_cycles 
;
; baudrate : the send baud rate
; io_pin   : the I/O pin which is used
;          : polarity == active_low  => 'direct' RS232
;          : polarity == active_high => inverting driver
; 
async_busy_send macro _baudrate, _stopbits, _io_pin
	local brt	= _baudrate
	local ipb 	= (xtal / 4) / _baudrate

	;local xyz
	cblock
		_file_1#v($)
		_file_2#v($)
	endc
	local send_shift_register = _file_1#v($)
	local bit_counter         = _file_2#v($)

	local loop
	local end_loop
	
	movwf send_shift_register	; save byte-to-send
	movlw (D'10'+_stopbits)		; incl. start & stopbit (10)
	movwf bit_counter		;	
	port_set_value _io_pin, off	; start bit
loop					; loop 9 times
	; loop delay = nr of instructions 
	;   +1  for the jump at the end
	;   -2  because delay_cycles is 2 more instructions
	;   +8 because the port_set_value calls are 9 cycles
	local x_loop_delay = (end_loop - loop) + D'7'
 	local loop_delay = D'17'
        local dc = ( ipb - loop_delay )
	delay_cycles dc

	btfsc send_shift_register, 0	;   set output 
	  port_set_value _io_pin, on	;   according to LSB
	btfss send_shift_register, 0	;
	  port_set_value _io_pin, off	;

	rrf send_shift_register, f	;   prepare next bit
	bsf send_shift_register, 7 ;   prepare stopbit
 
	decfsz bit_counter, f		; 
	goto loop			; end loop
end_loop
	endm


;
; Macro for 8/N/1 modulated asychronous serial 
; sending with busy delays.
; Must be instantiated with the choosen 
; io_pin, baudrate and modulation ferquency.
; Requires:
; - xtal must be defined (in KHz)
; - macro delay_cycles 
;
; baudrate  : the send baud rate
; io_pin    : the I/O pin which is used
;           : inactive during mark condition,
;           : pulsed active during space condition. 
; frequency : the modulation frequency in Hz
; 
async_modulated_busy_send macro _baudrate, _io_pin, _frequency, _duty_cycle
  local _ipb 	= ( xtal / 4 ) / _baudrate
  local _iph	= ( _duty_cycle * ( xtal / 4 ) ) / ( D'100' * _frequency )
  local _ipl	= ( ( xtal / 4 ) / _frequency ) - _iph
  local _pulses	= ( _ipb - D'10' ) / ( _iph + _ipl )
  local _extra	= _ipb - ( _pulses * ( _iph + _ipl ) )

  cblock
    _file_1#v($)
    _file_2#v($)
    _file_3#v($)
    _file_4#v($)
  endc
  local _send_shift_register = _file_1#v($)
  local _bit_counter         = _file_2#v($)
  local _pulse_counter       = _file_3#v($)
  local _bit_flag            = _file_4#v($)

  local _bit_loop
  local _pulse_loop
	
  movwf _send_shift_register	; save byte-to-send
  bcf _bit_flag, 0		; create the start-bit	

  movlw D'12'			; incl. start & stopbit (10)
  movwf _bit_counter		; loop for_each_bit
bit_loop

    movlw _pulses		;   loop for_each_pulse
    movwf _pulse_counter
pulse_loop

      port_set_value_from_inverted io_pin, bit_flag, 0
      delay_cycles ( _iph - D'9' )

      port_set_value io_pin, off
      delay_cycles ( _ipl - D'12' )

      decfsz _pulse_counter, f
      goto pulse_loop		;   end loop for_each_pulse

    delay_cycles _extra

    rrf _send_shift_register, f	; get next bit
    rlf _bit_flag, f		; into the flag
    bsf _send_shift_register, 7	; prepare the stopbits
 
    decfsz _bit_counter, f	;
    goto bit_loop		; end loop for_each_bit

	endm


;
; macro for 8/N/1 asynchronous receive
; using polling / busy waiting (no interrupts)
;
; macro arguments:
;	baudrate 	baudrate
;	port 		port which is used to receive
;	bit 		bit (within port) which is used to receive
;	rsr 		file register to contain the received byte
;	polarity 	1 for direct RS232, -1 for inverting buffer
;	waitflag	1 for wait-for-byte, 0 for return-on-idle
;
; The macro does not end with return, so it can be used either 
; in-line or (with a return after it) as a subroutine.
; xtal must be defined as the crystal frequency in Hz.
;
; example:
;		#define xtal D'4000000'
;		cblock
;			received_char
;		endc
;	get	async D'1200', porta, 0, received_char, 1, 1
;		return
;
; On exit the carry flag is set when a byte was received (this is
; always the case when waitflag==1). The carry is flag is clear when 
; waitflag==0 and the line was idle or a false start bit was detected. 
; No other errors are detected. Stop bit(s) are ignored, the macro 
; exits immediately after the 8'th bit.
;
; resources (in brackets: when used as subroutine)
;	code		15 (waitflag==0) or 16 (waitflag==1) [16, 17]
;	cycles		depends on baudrate and receive success,
;			3 [7] cycles when waitflag==0 and the line is idle
;	file 		2 local registers
;	stack		0 [1] entries 
;	w, flags	not preserved
;
; This receiver takes multiple samples per baud and uses a majority
; vote to determine the bit value. The number of samples depends on 
; the clock frequency and the baudrate. The maximum is 255 (must
; fit in one byte), the minimum depends on the required receive 
; quality (noise rejection and baudrate tolerance). Values below 3
; are rejected. The receive quality also depends on the rounding
; error which occurs. An error of 5e-3 is fine, and error of 30e-3 might
; still work. For waitflag==0 the receive quality depends also on the
; calling interval, which should be small compared to the baud length
; (a factor of 0.1 is good, 0.4 might work in some cases).
; 
; To achieve a lower baudrate than currently possible the clock 
; frequency can be lowered, or a delay can be introduced in the 
; sample_loop. To achieve a higher baudrate the clock frequency can be 
; raised, or the sample loop could be recoded (unwind the loop to for 
; instance 3 test/addlw pairs). When the code is modified the 
; nr_of_samples calculation must be adjusted (unless only nop's are added).
;
; The following standard baudrates can be realised without
; modifications (number of samples, rate error):
;      clock       lowest                  highest
;  frequency     baudrate                 baudrate
;     10 MHz         2400 (206, 0.27%)        38k4 (11, 1.70%)
;      4 MHz         1200 (165, 0.08%)        14k4 (12, 0.64%)
;      1 MHz          300 (165, 0.08%)        2400 (19, 0.16%)
;    455 KHz
;  32768 Hz                                    110 (13, 0.64%)
;
async_busy_receive macro _baudrate, _io_pin, _waitflag
	listing -1

	; check the arguments
	if ( _baudrate < 110 ) || ( _baudrate >115200 )
		error The supplied baud rate (#v(baudrate)) is illegal.
		messg - Valid values are 110..115200
	endif
	if ( _waitflag != 0 ) && ( _waitflag != 1 )
		error The supplied wait flag (#v(waitflag)) is illegal.
		messg - Valid values are 0 and 1
	endif

	local _same = 1			; destination = register file
	local _wreg = 0			; destination = w register
	local _user_error_detected	; used to force a fatal error

	local _start			; local lables
	local _bit_loop			;
	local _sample_loop		;
	local _sample_loop_end		;
	local _bit_loop_end		;
	local _done			;
	local _the_end			;

	local _ipb			; variables for the 
	local _nr_of_samples		; local calulations
	local _correction
	local _pivot			;
	local _error			;

	local _sample_cycles = 5	; timing assumptions
	local _bit_cycles = 10		; (verified at the end)

	; calculate instructions_per_baud
	_div_round _ipb, (xtal), 4 * _baudrate

	local _extra_delay = 0		; initial settings
	local _unroll = 0		;
	local _feasible = 0		;
		
		; calculate the number of samples
	_div_round _nr_of_samples, ( _ipb - _bit_cycles ), _sample_cycles
		
	while _nr_of_samples > 256	; must extra delay be added?
		_inc _sample_cycles
		_inc _extra_delay
		_div_round _nr_of_samples, ( _ipb - _bit_cycles ), _sample_cycles
	endw

	if _nr_of_samples < 3		; is unrolling necessary?
		_set _unroll, 1
		_set _sample_cycles, 2
		_div_round _nr_of_samples, ( _ipb - _bit_cycles ), _sample_cycles
	endif

		; calculate the addlw values to let the carry 
		; distinguish between mark and space
	_set _correction, D'255' - _nr_of_samples
	_div_round _pivot, _nr_of_samples, 2

		; calculate the error (in e-3)
	local _real_ipb = ( _nr_of_samples * _sample_cycles) + _bit_cycles
	local _temp = (D'10000' * (( _baudrate * 4 * _real_ipb) - (xtal))) / (xtal)
	_abs _temp
	_div_round _error, _temp, D'10'

		; verify the viability
	if _nr_of_samples > 255
		error The calculated number of samples 
		messg - (#v( _nr_of_samples)) does not fit in one byte.
	endif
	if _nr_of_samples < 3
		error The calculated number of samples 
		messg - (#v( _nr_of_samples)) is too low.
	endif
	if _error > 50
		error The calculated baud-rate error 
		messg - (#v( _error)*e-3) is too high.
	endif

	listing +1
	if 0 == 1 ; show the calculated values
		cycles in the inner (sample) loop	#v(_sample_cycles)
		(extra) cycles in the outer loop	#v(_bit_cycles)
		clock frequency				#v(xtal)
		baud_rate				#v(_baudrate)
		instructions per baud			#v(_ipb)
		samples per baud			#v(_nr_of_samples)
		correction factor			#v(_correction)
		pivot					#v(_pivot)
		real instructions per baud		#v(_real_ipb)
		error (* 1e-3)				#v(_error)
	endif

		; local variables
	cblock				
		_file1_#v($)
		_file2_#v($)
	endc
	local _sample_counter = _file1_#v($)
	local _rsr = _file2_#v($)
 
_start
		; handle idle line condition
	if _waitflag == 0
		clrc	
	endif
	if_port_value _io_pin, on	; if line is idle
	if _waitflag == 1
		goto _start		; try again
	else
		goto _the_end		; return immediately
	endif

	clrf _rsr			; clear rsr first, rsr is also 
					; used as bit_loop 'counter'

_bit_loop
	movlw _nr_of_samples		; iterate the sample loop 
	movwf _sample_counter		; nr_of_samples times

_sample_loop
	while _extra_delay > 0
		_dec _extra_delay
		nop
	endw
	if_port_value _io_pin, on	; count the spaces
		addlw -1		; w == #samples - #marks
	decfsz _sample_counter, _same
		goto _sample_loop
_sample_loop_end

	addlw _correction		; set carry iff
	addlw _pivot			; #spaces > #samples/2
					
	rrf _rsr, _same			; shift conclusion into rsr
					; initially rsr was zero,
					; start bit == 1, so
					; zero / no carry => false startbit
					; carry => 8 bits received
					; otherwise get next bit

	if _waitflag == 1
		skpnc			; if carry
			goto _done	; 	byte received

		skpnz			; otherwise, if zero
			goto _start	;	try again
	else
		skpc			; if carry goto done
			skpnz		; else if zero 
				goto _done	; goto done
	endif
	goto _bit_loop			; get next bit

_bit_loop_end
_done 
	comf _rsr, _wreg			; get correct result to w

_the_end
		; verify the timing assumptions
	local _sample_length = _sample_loop_end - _sample_loop
	;assert _sample_cycles == 1 + _extra_delay + _sample_length
	local _bit_length = _bit_loop_end - _bit_loop
	;assert _bit_cycles == 1 + _bit_length - _sample_length

 listing -1
 endm

 listing +1

