 list
 ;---------------------------------------------------------------------
 ; CONTROL:
 ;	file		misc.inc
 ;	author		wouter van ooijen <wf@xs4all.nl>
 ; 	version		1.0
 ; 	date		05-09-1997
 ; 	language	MPASM assembler for Microchip PIC processors
 ;      summary         miscallaneous small tools
 ;      info            http://www.xs4all.nl/~wf/wouter/pic/wloader
 ;---------------------------------------------------------------------
 nolist
 ; Copyright (C) 2000 Wouter van Ooijen
 ;
 ; This application is free software; you can redistribute it and/or
 ; modify it under the terms of the GNU Library General Public
 ; License as published by the Free Software Foundation; either
 ; version 2 of the License, or (at your option) any later version.
 ;
 ; This library is distributed in the hope that it will be useful,
 ; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ; Library General Public License for more details.
 ; 
 ; You should have received a copy of the GNU Library General Public
 ; License along with this library; if not, write to the
 ; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 ; Boston, MA  02111-1307, USA.
 ;---------------------------------------------------------------------
	
 nolist

; DESCRIPTION:
; _if, ...	obfusciated MPASM directives
; _set, ... 	macros to avoid left-margin labels
; listing 	nestable alternative for list/nolist
; assert	documents and verifies assumptions
; uses		intelligent #include
; true, ...	boolean type
; 
; _div		various divide & remainder macro's
;
; REQUIRES:
;	nothing
;
; RESOURCES (once):
; 	code		0
; 	file bytes	0
;

; Mhz			; D'1000000'
; Khz			; D'1000'
#define MHz D'1000000'
#define KHz D'1000'

;
; SUMMARY:
;	obfusciated MPASM directives
;
; DESCRIPTION:
; 	Use these alternatives for the plain MPASM directives 
;	when the directive line must be subject to macro expansion. 
;	A line with a plain directive is not subject to macro expansion. 
;	Sadly it is not possible to obfusciate the ENDM directive.
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
;
; 
#define _include	#include
#define _define		#define
#define _undefine	#undefine
#define _macro		macro
#define _ifdef		#ifdef
#define _ifndef		#ifndef
#define _else		#else
#define _endif		#endif


;
; SUMMARY:
;	various set-value macros
;
; DESCRIPTION:
; 	Use these macros to set macro variables without using
;	a label at the left margin. I don't like non-code labels 
;	at the left margin, beacuse that breaks the normal
;	indettation. 
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
;
; like set, but aligns properly
_set	macro name, value
name	set (value)
	endm
; like equ, but aligns properly
_equ	macro name, value
name	equ (value)
	endm
; like +=, but aligns properly
_add	macro name, value
	_set name, name + (value)
	endm
; like ++, but aligns properly
_inc	macro name
	_set name, (name+1)
	endm
; like --, but can be aligned properly
_dec	macro name
	_set name, (name-1)
	endm
; set name to maximum of current and supplied value
_max 	macro name, value
	if value > name 
		_set name, value
	endif
	endm
; set name to its absolute value
_abs macro name
	if name < 0
		_set name, -name
	endif
	endm


;
; SUMMARY:
;	listing: nestable alternative for nolist/list
;
; DESCRIPTION:
; 	Use these macros to switch listing on and off in a
;	nestable way. list is on iff sum(delta) > 0.
;	It would be nice if the 'list' which switches
;	the listing on would not show on the listing...
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
	_set _listlevel, 0
listing macro delta
 ;	nolist
 list
	_add _listlevel, delta
	; messg listlevel d #v(delta) is now #v(_listlevel) ; for debug
	if _listlevel < 0
 		exitm
	endif
 list;ignore
	endm


;
; SUMMARY:
;	assert: documents and verifies assumptions 
;
; DESCRIPTION:
; 	Use this macro to document and verify your assumptions.
;	This will help the reader of your code, and it will
;	prevent the use of your code in situations where
;	the assumption does not hold.
;
;	The macro force_assembley_failure prevents that messg's
;	in an otherwise succesfull assembley would go unnoticed.
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
;
assert macro condition
	nolist
	listing -1
	if !(condition)
		error "Assertion failure"
	endif
	listing +1
	endm


;
; SUMMARY:
;	uses: intellingent #include
;
; DESCRIPTION:
; 	Use this macro includes the specified file when
;	it has not been 'used' already. 
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
;
with macro x
	nolist
	listing -1
	_ifndef x#v(0)_included
		_include x
		_define x#v(0)_included 
	#endif
	listing +1
	endm


;
; SUMMARY:
;	true, ... : boolean type
;
; DESCRIPTION:
; 	true and false are the boolean constants.
;	invert( true ) == false and
;	invert( false ) == true
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
;
	#define true 1
	#define false -1
	#define invert(arg) (-1*(arg))


;
; SUMMARY:
;	_div	various divide and remainder macros
;
; DESCRIPTION:
; 	_div		chopping divide (/)
;	_div_rem	divide & remainder (positive remainder)
;	_div_round	rounding divide
;	_div_rem_round	rounding divide & remainder 
;
; RESOURCES (each use):
;	none
; RESOURCES (once):
;	none
;
_div_rem macro q, r, value, divident
	local x = value
	local y = divident
	assert ( (x) >= 0 && (y) > 0 )
	_set q, (x) / (y)
	_set r, (x) - ((y) * q)
	assert ( r >=0 && r < (y) && (x) == ((y) * q) + r )
	endm
_div macro q, value, divident
	nolist
	listing -1
	local dummy
	_div_rem q, dummy, value, divident
	listing +1
	endm
_mod macro r, value, divident
	nolist
	listing -1
	local dummy
	_div_rem dummy, r, value, divident
	listing +1
	endm
_div_round macro q, value, divident
	nolist
	listing -1
	local x = value
	local y = divident
	assert ( x >= 0 && y > 0 )
	_set q, (x + (y / 2)) / y
	assert (2*(x-(y*q))>=-y)
	assert (2*(x-(y*q))<y)
	listing +1
	endm
_div_rem_round macro q, r, value, divident
	nolist
	listing -1
	local x = value
	local y = divident
	assert (x >= 0 && y > 0)
	_set q, (x + (y / 2 )) / y
	_set r, x - (y * q)
	assert (r >= 0)
	assert (r < y) 
	assert (x == ((y * q) + r))
	listing +1
	endm

; move macro's
move_from_via_w_to macro _x, _y
 movfw _x
 movwf _y
 endm

exchange_via_w macro _x, _y
 movfw _x
 xorwf _y, w
 xorwf _x, f
 xorwf _y, f
 endm

move_const_via_w_to macro _x, _y
 movlw _x
 movwf _y
 endm


